from octostar.client import make_client, UserContext, User
from streamlit_octostar_research.desktop import whoami
from streamlit.runtime.scriptrunner import get_script_run_ctx
import hashlib
import streamlit as st

def as_running_user(force_refresh=False):
    script_ctx = get_script_run_ctx()
    internal_st_key = "__run_user"
    prev_run_id = st.session_state[internal_st_key].get('prev_run_id')
    prev_user = st.session_state[internal_st_key].get('prev_user')
    is_first_time = False
    if prev_run_id != id(script_ctx.script_requests): # this changes at every st rerun
        is_first_time = True
    st.session_state[internal_st_key]['prev_run_id'] = id(script_ctx.script_requests)
    if not is_first_time and not force_refresh and prev_user:
        return UserContext(prev_user)
    running_user = whoami()
    if not running_user:
        if prev_user:
            return UserContext(prev_user)
        else:
            st.stop()
    running_user_hash = hashlib.md5(running_user['os_jwt'].encode('utf-8'))
    if not prev_user or hash(prev_user) != running_user_hash:
        client = make_client(fixed_jwt=running_user['os_jwt'])
        user = User(client)
        st.session_state[internal_st_key]['prev_user'] = user
    return UserContext(user)


import streamlit as st

from streamlit_octostar_research.desktop import get_open_workspace_ids
from octostar.utils.ontology import query_ontology
from octostar.utils.workspace.permissions import get_permissions, PermissionLevel

def get_workspaces(prev_workspaces=None, include_closed=False):
    if not include_closed:
        workspace_ids = get_open_workspace_ids()
    else:
        workspace_ids = [w['os_entity_uid'] for w in query_ontology.sync(f"SELECT `os_entity_uid` FROM `dtimbr`.`os_workspace`")]
    if not workspace_ids and prev_workspaces is not None:
        return prev_workspaces
    if not workspace_ids:
        st.stop()
    workspace_ids = ", ".join(["'" + workspace_id + "'" for workspace_id in workspace_ids])
    workspaces = query_ontology.sync(f"SELECT `entity_label`, `os_entity_uid` FROM `dtimbr`.`os_workspace` WHERE `os_entity_uid` IN ({workspace_ids})")
    workspaces_permissions = get_permissions.sync([w['os_entity_uid'] for w in workspaces])
    workspaces = {workspace['os_entity_uid']: {**workspace, 'os_permission': workspaces_permissions.get(workspace['os_entity_uid'], PermissionLevel.NONE)} for workspace in workspaces}
    return dict(filter(lambda x: x[1]['os_permission'] > PermissionLevel.NONE, workspaces.items()))


from typing import Union, List, Literal, Any, Callable, Optional
import json
import hashlib

def recursive_update_dict(
    d1: dict,
    d2: dict,
    op: Callable[[Any, Any], Any],
    recurse: bool = True
) -> dict:
    if not isinstance(d1, dict) or not isinstance(d2, dict):
        if d1 is None:
            return d2
        elif d2 is None:
            return d1
        else:
            return op(d1, d2)
    for k2, v2 in d2.items():
        if isinstance(v2, dict):
            v1 = d1.get(k2, {})
            if isinstance(v1, dict):
                if recurse:
                    if isinstance(recurse, int):
                        recurse -= 1
                    d1[k2] = recursive_update_dict(v1, v2, op, recurse)
                else:
                    d1[k2] = op(v1, v2)
            else:
                if k2 in d1:
                    d1[k2] = op(v1, v2)
                else:
                    d1[k2] = v2
        else:
            if k2 in d1:
                d1[k2] = op(d1[k2], v2)
            else:
                d1[k2] = v2
    return d1

def travel_dict(
    d: dict,
    keylist: Optional[List[str]] = None,
    mode: Literal['r', 'w'] = 'r',
    strict: bool = False
) -> Union[Any, callable]:
    def __set(pointer, key, value):
        pointer[key] = value
    pointer = d
    if not keylist:
        keylist = []
    match mode:
        case 'w':
            for key in keylist[:-1]:
                if key not in pointer:
                    if strict:
                        raise KeyError(key)
                    pointer[key] = dict()
                pointer = pointer[key]
            return lambda v: __set(pointer, key, v)
        case 'r':
            for key in keylist:
                if not hasattr(pointer, "__contains__"):
                    if strict:
                        raise KeyError(key)
                    return pointer
                else:
                    if key not in pointer:
                        if strict:
                            raise KeyError(key)
                        return pointer
                    else:
                        pointer = pointer[key]
            return pointer                 

def jsondict_hash(d: dict):
    def _make_hash(o):
        if isinstance(o, list):
            return hash([_make_hash(e) for e in o])
        elif isinstance(o, dict):
            new_o = {k: _make_hash(v) for k, v in sorted(o.items())}
            return hash(tuple(frozenset(new_o.items())))
        else:
            return hash(o)
    dict_str = json.dumps(d, sort_keys=True, default=_make_hash)
    return hashlib.md5(dict_str.encode()).hexdigest()